from typing import Any, Dict, Generator, Generic, List, Optional, Type, Union

from denial import InnerNoneType
from printo import descript_data_object

from pristan.common_types import (
    PluginResult,
    SlotFunction,
    SlotPapameters,
    SlotResult,
)
from pristan.components.plugin import Plugin
from pristan.components.plugins_group import PluginsGroup
from pristan.components.slot_code_representer import SlotCodeRepresenter
from pristan.components.slot_code_representer import sentinel as return_type_sentinel


class SlotCaller(Generic[PluginResult]):
    # TODO: consider to delete this "type: ignore" if python 3.8 deleted from the matrix
    def __init__(self, code_representation: SlotCodeRepresenter, slot_name: Optional[str], slot_function: SlotFunction[SlotPapameters, SlotResult[PluginResult]], type_check: bool) -> None:  # type: ignore[type-arg, unused-ignore]
        self.code_representation = code_representation
        self.slot_name = slot_name
        self.slot_function = slot_function
        self.type_check = type_check

    def __repr__(self) -> str:
        return descript_data_object(
            type(self).__name__,
            [],
            {
                'code_representation': self.code_representation,
                'slot_name': self.slot_name,
                'slot_function': self.slot_function,
                'type_check': self.type_check,
            },
        )

    def __call__(self, plugins: Union[PluginsGroup[PluginResult], List[Plugin[PluginResult]]], *args: SlotPapameters.args, **kwargs: SlotPapameters.kwargs) -> SlotResult[PluginResult]:  # type: ignore[return]
        if not self.code_representation.is_empty and not plugins:
            if self.code_representation.returns_list:
                if self.code_representation.returning_type is return_type_sentinel:
                    returns_type: Union[Type[Any], InnerNoneType] = List
                else:
                    returns_type = List[self.code_representation.returning_type]  # type: ignore[name-defined]
            elif self.code_representation.returns_dict:
                if self.code_representation.returning_type is return_type_sentinel:
                    returns_type = Dict
                else:
                    returns_type = Dict[str, self.code_representation.returning_type]  # type: ignore[name-defined]
            else:
                returns_type = self.code_representation.returning_type

            # TODO: consider to delete this "type: ignore" if python 3.9 deleted from the matrix
            result: SlotResult[PluginResult] = Plugin(self.slot_name if self.slot_name is not None else self.slot_function.__name__, self.slot_function, returns_type, self.type_check, False)(*args, **kwargs)  # type: ignore[assignment, unused-ignore]

            if self.code_representation.returning_type is return_type_sentinel and not self.code_representation.returns_dict and not self.code_representation.returns_list:
                result = None

            return result

        if self.code_representation.returns_list:
            return [plugin(*args, **kwargs) for plugin in plugins]

        if self.code_representation.returns_dict:
            return {plugin.name: plugin(*args, **kwargs) for plugin in plugins}

        for plugin in plugins:
            plugin(*args, **kwargs)


class CallerWithPlugins(Generic[PluginResult]):
    def __init__(self, caller: SlotCaller[PluginResult], plugins: List[Plugin[PluginResult]]) -> None:
        self.caller = caller
        self.plugins = plugins

    def __call__(self, *args: SlotPapameters.args, **kwargs: SlotPapameters.kwargs) -> SlotResult[PluginResult]:
        return self.caller(self.plugins, *args, **kwargs)

    def __repr__(self) -> str:
        return descript_data_object(
            type(self).__name__,
            [],
            {
                'caller': self.caller,
                'plugins': self.plugins,
            },
        )

    def __iter__(self) -> Generator[Plugin[PluginResult], None, None]:
        yield from self.plugins

    def __bool__(self) -> bool:
        return bool(self.plugins)

    def __len__(self) -> int:
        return len(self.plugins)



from functools import partial, wraps
from typing import Callable, Optional, Union, overload

from pristan.common_types import PluginResult, SlotFunction, SlotPapameters
from pristan.components.slot import Slot


# TODO: concider to delete all "type: ignore" comments with "unused-ignore"'s in this file if python 3.8 deleted from the matrix
@overload
def slot(function: SlotFunction[SlotPapameters, PluginResult], /) -> SlotFunction[SlotPapameters, PluginResult]: ...  # type: ignore[type-arg, unused-ignore] # pragma: no branch

@overload
def slot(*, signature: Optional[str] = None, name: Optional[str] = None, max_plugins: Optional[int] = None, type_check: bool = True, entrypoint_group: str = 'pristan') -> Callable[[SlotFunction[SlotPapameters, PluginResult]], SlotFunction[SlotPapameters, PluginResult]]: ...  # type: ignore[type-arg, unused-ignore] # pragma: no branch

@overload
def slot(function: str, *, signature: Optional[str] = None, name: Optional[str] = None, max_plugins: Optional[int] = None, type_check: bool = True, entrypoint_group: str = 'pristan') -> Callable[[SlotFunction[SlotPapameters, PluginResult]], SlotFunction[SlotPapameters, PluginResult]]: ...  # type: ignore[type-arg, unused-ignore] # pragma: no branch

def slot(function: Optional[Union[SlotFunction[SlotPapameters, PluginResult], str]] = None, /, *, signature: Optional[str] = None, name: Optional[str] = None, max_plugins: Optional[int] = None, type_check: bool = True, entrypoint_group: str = 'pristan') -> Union[SlotFunction[SlotPapameters, PluginResult], Callable[[SlotFunction[SlotPapameters, PluginResult]], SlotFunction[SlotPapameters, PluginResult]]]:  # type: ignore[misc, type-arg, unused-ignore] # noqa: PLR0913
    if callable(function):
        return wraps(function)(Slot(function, signature, name, max_plugins, type_check, entrypoint_group))  # type: ignore[arg-type, return-value, unused-ignore]

    if isinstance(function, str):
        if name is not None and name != function:
            raise ValueError('You have specified two different names for the slot.')
        name = function

    return partial(slot, signature=signature, name=name, max_plugins=max_plugins, type_check=type_check, entrypoint_group=entrypoint_group)  # type: ignore[arg-type, unused-ignore]



from typing import Any, Generic, Type, Union

from denial import InnerNoneType
from printo import descript_data_object
from simtypes import check

from pristan.common_types import PluginFunction, PluginResult, SlotPapameters
from pristan.components.slot_code_representer import sentinel as return_type_sentinel


class Plugin(Generic[PluginResult]):
    # TODO: consider to delete this "type: ignore" if python 3.9 deleted from the matrix
    def __init__(self, name: str, plugin_function: PluginFunction[SlotPapameters, PluginResult], expected_result_type: Union[InnerNoneType, Type[Any]], type_check: bool, unique: bool) -> None:  # type: ignore[type-arg, unused-ignore]
        self.plugin_function = plugin_function
        self.requested_name = name
        self.name = name
        self.expected_result_type = expected_result_type
        self.type_check = type_check
        self.unique = unique

    def __repr__(self) -> str:
        return descript_data_object(
            type(self).__name__,
            (self.name,),
            {
                'plugin_function': self.plugin_function,
                'expected_result_type': self.expected_result_type,
                'type_check': self.type_check,
                'unique': self.unique,
            },
        )

    def __call__(self, *args: SlotPapameters.args, **kwargs: SlotPapameters.kwargs) -> PluginResult:
        # TODO: try to delete this "type: ignore" comments if python 3.8 deleted from CI
        result = self.plugin_function(*args, **kwargs)  # type: ignore[arg-type, unused-ignore]

        if self.type_check and self.expected_result_type is not return_type_sentinel and not check(result, self.expected_result_type, strict=True):  # type: ignore[arg-type]
            raise TypeError(f'The type {type(result).__name__} of the plugin\'s "{self.name}" return value {result!r} does not match the expected type {self._get_class_name(self.expected_result_type)}.')  # type: ignore[union-attr, unused-ignore]

        return result  # type: ignore[no-any-return, unused-ignore]

    def set_name(self, name: str) -> None:
        self.name = name

    @staticmethod
    def _get_class_name(_type: Any) -> str:
        try:
            return _type.__name__  # type: ignore[no-any-return]
        except AttributeError:  # pragma: no cover
            return str(_type)



from ast import Constant, Expr, Pass, Return, parse
from ast import Dict as ASTDict
from ast import List as ASTList
from functools import cached_property
from importlib.metadata import PackageNotFoundError, version
from inspect import getmodule
from typing import (
    Any,
    Callable,
    Dict,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from denial import InnerNoneType
from getsources import getclearsource
from packaging.version import Version
from printo import descript_data_object

from pristan.errors import CannotGetVersionsError

sentinel = InnerNoneType()

class SlotCodeRepresenter:
    def __init__(self, function: Callable[..., Any]) -> None:
        self.function = function
        self.returning_type  # noqa: B018

    def __repr__(self) -> str:
        return descript_data_object(type(self).__name__, [self.function], {})

    @cached_property
    def base_module(self) -> Optional[str]:
        module = getmodule(self.function)
        if module is not None:
            return module.__name__.split('.')[0]
        return None  # pragma: no cover

    @cached_property
    def package_version(self) -> Optional[Version]:
        if self.base_module is None:
            return None  # pragma: no cover
        try:
            version_identifier = version(self.base_module)
            return Version(version_identifier)
        except PackageNotFoundError:
            return None

    @cached_property
    def returning_type(self) -> Union[InnerNoneType, Type[Any]]:
        hints = get_type_hints(self.function)
        return_hint = hints.get('return', sentinel)

        if return_hint is sentinel:
            return sentinel

        if list in (return_hint, get_origin(return_hint)):
            args = get_args(return_hint)
            if args:
                return args[0]  # type: ignore[no-any-return]
            return sentinel

        if dict in (return_hint, get_origin(return_hint)):
            args = get_args(return_hint)
            if args:
                if args[0] is not str or len(args) != 2:
                    raise TypeError('Incorrect type annotation for the dict.')
                return args[1]  # type: ignore[no-any-return]
            return sentinel

        return return_hint  # type: ignore[no-any-return]

    @cached_property
    def returns_list(self) -> bool:
        hints = get_type_hints(self.function)
        return_hint = hints.get('return', sentinel)

        if return_hint is sentinel:
            return False

        return list in (return_hint, get_origin(return_hint))

    @cached_property
    def returns_dict(self) -> bool:
        hints = get_type_hints(self.function)
        return_hint = hints.get('return', sentinel)

        if return_hint is sentinel:
            return False

        return dict in (return_hint, get_origin(return_hint))

    @cached_property
    def is_empty(self) -> bool:
        converted_source_code = getclearsource(self.function)

        tree = parse(converted_source_code)
        body = tree.body[0].body  # type: ignore[attr-defined]

        return all(
            (
                isinstance(body_statement, Pass) or (isinstance(body_statement, Expr) and isinstance(body_statement.value, Constant) and (body_statement.value.value is Ellipsis or isinstance(body_statement.value.value, str)))
                for body_statement in body
            ),
        ) or (
            len(body) == 1 and isinstance(body[0], Return)
            and (
                (self.returns_list and isinstance(body[0].value, ASTList) and not body[0].value.elts)
                or ((self.returns_dict and isinstance(body[0].value, ASTDict) and not body[0].value.keys and not body[0].value.values))
            )
        )

    def check_package_version(self, expression: Optional[str]) -> bool:
        if expression is None:
            return True

        if self.package_version is None:
            raise CannotGetVersionsError('It is not possible to obtain the name of the package in which the slot is declared.')

        expression = expression.strip()

        start_signs: Dict[str, Callable[[str], bool]] = {
            '==': lambda x: self.package_version == Version(x),
            '>=': lambda x: self.package_version >= Version(x),  # type: ignore[operator]
            '<=': lambda x: self.package_version <= Version(x),  # type: ignore[operator]
            '>': lambda x: self.package_version > Version(x),  # type: ignore[operator]
            '<': lambda x: self.package_version < Version(x),  # type: ignore[operator]
        }

        for start, checker in start_signs.items():
            if expression.startswith(start):
                if checker(expression[len(start):]):
                    return True

        return False



from collections import defaultdict
from typing import Any, DefaultDict, Generator, Generic, List, Optional

from printo import descript_data_object

from pristan.common_types import PluginResult
from pristan.components.plugin import Plugin


class PluginsGroup(Generic[PluginResult]):
    """
    The class is a collection of plugins.

    All collection operations are totally not thread-safe.
    """

    def __init__(self, caller: 'SlotCaller', plugins: Optional[List[Plugin[PluginResult]]] = None) -> None:  # type: ignore[name-defined] # noqa: F821
        self.caller = caller
        self.plugins: List[Plugin[PluginResult]] = []
        self.plugins_by_requested_names: DefaultDict[str, List[Plugin[PluginResult]]] = defaultdict(list)

        if plugins:
            self.add(*plugins)

    def __repr__(self) -> str:
        return descript_data_object(type(self).__name__, [self.caller], {'plugins': self.plugins}, filters={'plugins': lambda x: bool(self.plugins)})  # noqa: ARG005

    def __bool__(self) -> bool:
        return bool(self.plugins)

    def __len__(self) -> int:
        return len(self.plugins)

    def __iter__(self) -> Generator[Plugin[PluginResult], None, None]:
        yield from self.plugins

    def __contains__(self, item: Any) -> bool:
        from pristan.components.plugin import Plugin  # noqa: PLC0415

        if isinstance(item, str):
            if item.isidentifier():
                return item in self.plugins_by_requested_names
            if self._is_identifier_with_number(item):
                splitted = item.split('-')
                first_part = splitted[0]
                second_part = splitted[1]
                if second_part == '1':
                    return first_part in self
                if first_part not in self.plugins_by_requested_names:
                    return False
                for plugin in self.plugins_by_requested_names[first_part]:
                    if plugin.name == item:
                        return True
                return False
            raise ValueError(f'The plugin name string must look like either a valid Python identifier or an identifier plus one or more digits separated by a hyphen, for example, “name-22”. "{item}" is not a valid name for a plugin.')

        if isinstance(item, Plugin):
            return item.requested_name in self.plugins_by_requested_names and any(x.name == item.name for x in self.plugins_by_requested_names[item.requested_name])  # pragma: no branch

        raise TypeError('Checking for inclusion is only possible for strings of a valid format or for plugin objects.')

    def __getitem__(self, key: str) -> 'CallerWithPlugins':  # type: ignore[name-defined] # noqa: F821
        from pristan.components.slot_caller import CallerWithPlugins  # noqa: PLC0415

        if isinstance(key, str):
            if key.isidentifier():
                if key in self.plugins_by_requested_names:
                    return CallerWithPlugins(self.caller, [x for x in self.plugins_by_requested_names[key]])
                return CallerWithPlugins(self.caller, [])

            if self._is_identifier_with_number(key):
                splitted = key.split('-')
                first_part = splitted[0]
                second_part = splitted[1]
                if first_part not in self.plugins_by_requested_names:
                    return CallerWithPlugins(self.caller, [])
                for plugin in self.plugins_by_requested_names[first_part]:
                    if (second_part == '1' and plugin.name == first_part) or plugin.name == key:
                        return CallerWithPlugins(self.caller, [plugin])
                return CallerWithPlugins(self.caller, [])

        raise KeyError('You have used an invalid key. Strings that are suitable as keys are valid Python identifiers, or the same strings with a number separated by a hyphen (e.g., "a", "a-5").')

    @staticmethod
    def _is_identifier_with_number(name: str) -> bool:
        return '-' in name and (lambda parts: len(parts) == 2 and parts[0].isidentifier() and parts[1].isdigit() and parts[1] != '0')(name.split('-'))  # noqa: PLC3002

    def add(self, *plugins: 'Plugin[PluginResult]') -> None:
        for plugin in plugins:
            self.plugins.append(plugin)
            self.plugins_by_requested_names[plugin.requested_name].append(plugin)

    def delete_last_by_name(self, name: str) -> None:
        self.plugins_by_requested_names[name].pop()
        self.plugins.pop()



from typing import Callable, Dict, List, Optional, TypeVar, Union

try:
    # TODO: ParamSpec appeared with python 3.10, so delete this try-except if python 3.9 deleted from the matrix
    from typing import ParamSpec  # type: ignore[attr-defined, unused-ignore]
except ImportError:  # pragma: no cover
    from typing_extensions import ParamSpec  # type: ignore[assignment, unused-ignore]

SlotPapameters = ParamSpec('SlotPapameters')
PluginResult = TypeVar('PluginResult')
SlotResult = Optional[Union[List[PluginResult], Dict[str, PluginResult]]]
SlotFunction = Callable[SlotPapameters, Optional[Union[List[PluginResult], Dict[str, PluginResult]]]]  # type: ignore[valid-type, misc, unused-ignore]
PluginFunction = Callable[SlotPapameters, PluginResult]  # type: ignore[valid-type, misc, unused-ignore]



try:
    from importlib_metadata import (  # type: ignore[import-not-found, unused-ignore]
        entry_points,  # type: ignore[import-not-found, unused-ignore]
    )
except ImportError:  # type: ignore[assignment, unused-ignore] # pragma: no cover
    from importlib.metadata import (  # type: ignore[assignment, unused-ignore]
        entry_points,  # type: ignore[assignment, unused-ignore]
    )

from threading import RLock
from typing import (
    Any,
    Callable,
    Generator,
    Generic,
    Optional,
    Tuple,
    Union,
    overload,
)

from printo import descript_data_object, not_none
from sigmatch import PossibleCallMatcher
from sigmatch.errors import SignatureMismatchError

from pristan.common_types import (
    PluginFunction,
    PluginResult,
    SlotFunction,
    SlotPapameters,
    SlotResult,
)
from pristan.components.plugin import Plugin
from pristan.components.plugins_group import PluginsGroup
from pristan.components.slot_caller import CallerWithPlugins, SlotCaller
from pristan.components.slot_code_representer import SlotCodeRepresenter
from pristan.components.slot_code_representer import sentinel as return_type_sentinel
from pristan.errors import (
    PrimadonnaPluginError,
    StrangeTypeAnnotationError,
    TooManyPluginsError,
)


# TODO: consider to delete all the "type: ignore"d comments if python 3.9 deleted from the matrix
class Slot(Generic[PluginResult]):
    def __init__(self, slot_function: SlotFunction[SlotPapameters, SlotResult[PluginResult]], signature: Optional[str], slot_name: Optional[str], max_plugins: Optional[int], type_check: bool, entrypoint_group: str) -> None:  # type: ignore[type-arg, unused-ignore] # noqa: PLR0913
        if max_plugins is not None and max_plugins < 0:
            raise ValueError('The maximum number of plugins cannot be less than zero.')

        self.slot_function = slot_function
        self.code_representation = SlotCodeRepresenter(self.slot_function)

        if not self.code_representation.returns_list and not self.code_representation.returns_dict and self.code_representation.returning_type is not return_type_sentinel:
            raise StrangeTypeAnnotationError('The return type annotation for a slot must be either a list or a dict, or remain empty.')

        self.signature = signature
        self.slot_name = slot_name
        self.slot_function = slot_function
        self.max_number_of_plugins = max_plugins
        self.type_check = type_check
        self.entrypoint_group = entrypoint_group

        self.lock = RLock()

        self.caller: SlotCaller[PluginResult] = SlotCaller(self.code_representation, self.slot_name, self.slot_function, self.type_check)
        self.plugins: PluginsGroup[PluginResult] = PluginsGroup(self.caller)
        self.backed_caller = CallerWithPlugins(self.caller, self.plugins.plugins)

        self._compare_signatures(self.slot_function, self.slot_function)  # type: ignore[arg-type, unused-ignore]

        self.loaded = False

    def __call__(self, *args: SlotPapameters.args, **kwargs: SlotPapameters.kwargs) -> SlotResult[PluginResult]:
        self._load_entrypoints()
        return self.backed_caller(*args, **kwargs)

    def __iter__(self) -> Generator[Plugin[PluginResult], None, None]:
        self._load_entrypoints()
        yield from self.plugins

    def __getitem__(self, key: str) -> CallerWithPlugins[PluginResult]:
        self._load_entrypoints()
        return self.plugins[key]  # type: ignore[no-any-return]

    def __repr__(self) -> str:
        return descript_data_object(
            type(self).__name__,
            [self.slot_function],
            {
                'signature': self.signature,
                'slot_name': self.slot_name,
                'max_plugins': self.max_number_of_plugins,
                'type_check': self.type_check,
            },
            filters={
                'signature': not_none,
                'slot_name': not_none,
                'max_plugins': not_none,
                'type_check': lambda x: x != True,
            },
        )

    def __contains__(self, item: Any) -> bool:
        return item in self.plugins

    def __len__(self) -> int:
        return len(self.plugins)

    @overload
    def plugin(self, plugin_function_or_name: Optional[str], unique: bool = False) -> Callable[[PluginFunction[SlotPapameters, PluginResult]], PluginFunction[SlotPapameters, PluginResult]]:  # type: ignore[type-arg, unused-ignore]
        ...  # pragma: no cover

    @overload
    def plugin(self, plugin_function_or_name: PluginFunction[SlotPapameters, PluginResult], unique: bool = False) -> PluginFunction[SlotPapameters, PluginResult]:  # type: ignore[type-arg, unused-ignore]
        ...  # pragma: no cover

    def plugin(self, plugin_function_or_name: Optional[Union[PluginFunction[SlotPapameters, PluginResult], str]] = None, unique: bool = False, engine: Optional[str] = None) -> Union[Callable[[PluginFunction[SlotPapameters, PluginResult]], PluginFunction[SlotPapameters, PluginResult]], PluginFunction[SlotPapameters, PluginResult]]:  # type: ignore[type-arg, unused-ignore]
        if isinstance(plugin_function_or_name, str):
            if not plugin_function_or_name.isidentifier():
                raise ValueError('The plugin name must be a valid Python identifier.')
            get_plugin_name: Callable[[PluginFunction[SlotPapameters, PluginResult]], str] = lambda function: plugin_function_or_name  # type: ignore[type-arg, unused-ignore] # noqa: E731, ARG005

        elif callable(plugin_function_or_name):
            get_plugin_name = lambda function: plugin_function_or_name.__name__  # noqa: E731, ARG005

        elif plugin_function_or_name is None:
            get_plugin_name = lambda function: function.__name__  # noqa: E731

        else:
            raise TypeError('Only a function or plugin name followed by a function can be passed to the decorator.')

        def decorator(plugin_function: PluginFunction[SlotPapameters, PluginResult]) -> PluginFunction[SlotPapameters, PluginResult]:  # type: ignore[type-arg, unused-ignore]
            # TODO: consider to delete this "type: ignore" if python 3.8 deleted from the matrix
            self._compare_signatures(self.slot_function, plugin_function)  # type: ignore[arg-type, unused-ignore]
            self._add_plugin(get_plugin_name(plugin_function), plugin_function, unique, engine)
            return plugin_function

        if plugin_function_or_name is None or isinstance(plugin_function_or_name, str):
            return decorator

        return decorator(plugin_function_or_name)

    def keys(self) -> Tuple[str, ...]:
        self._load_entrypoints()
        return tuple(self.plugins.plugins_by_requested_names.keys())

    def _load_entrypoints(self) -> None:
        with self.lock:
            if not self.loaded:
                for point in entry_points(group=self.entrypoint_group):
                    point.load()
                self.loaded = True

    def _add_plugin(self, name: str, function: PluginFunction[SlotPapameters, PluginResult], unique: bool, engine: Optional[str]) -> None:  # type: ignore[type-arg, unused-ignore]
        plugin: Plugin = Plugin(name, function, self.code_representation.returning_type, self.type_check, unique)  # type: ignore[type-arg]

        with self.lock:
            if len(self.plugins) == self.max_number_of_plugins:
                raise TooManyPluginsError(f'The maximum number of plugins for this slot is {self.max_number_of_plugins}.')

            if self.code_representation.check_package_version(engine):
                self.plugins.add(plugin)
                if len(self.plugins.plugins_by_requested_names[name]) > 1:
                    plugin.set_name(f'{name}-{len(self.plugins.plugins_by_requested_names[name])}')
                    for other_plugin in self.plugins.plugins_by_requested_names[name]:
                        if other_plugin.unique:
                            self.plugins.delete_last_by_name(name)
                            raise PrimadonnaPluginError(f'Plugin "{other_plugin.name}" claims to be unique, but there are other plugins with the same name.')

    def _compare_signatures(self, slot_function: SlotFunction[SlotPapameters, SlotResult[PluginResult]], plugin_function: PluginFunction[SlotPapameters, PluginResult]) -> None:  # type: ignore[type-arg, unused-ignore]
        if self.signature is not None:
            PossibleCallMatcher(self.signature).match(plugin_function, raise_exception=True)
        elif not PossibleCallMatcher.from_callable(slot_function) & PossibleCallMatcher.from_callable(plugin_function):
            raise SignatureMismatchError('No common calling method has been found between the slot and the plugin.')
